""" Generators for basis evaluations """

from dune.perftool.generation import (cached,
                                      class_member,
                                      domain,
                                      generator_factory,
                                      iname,
                                      include_file,
                                      instruction,
                                      preamble,
                                      symbol,
                                      temporary_variable,
                                      )
from dune.perftool.pdelab.quadrature import (name_quadrature_position_in_cell,
                                             quadrature_iname,
                                             )
from dune.perftool.pdelab.geometry import (dimension_iname,
                                           name_dimension,
                                           name_jacobian_inverse_transposed,
                                           to_cell_coordinates,
                                           )
from dune.perftool.pdelab.localoperator import (lop_template_ansatz_gfs,
                                                lop_template_test_gfs,
                                                )
from dune.perftool.pdelab.driver import FEM_name_mangling
from dune.perftool.pdelab import restricted_name
from pymbolic.primitives import Product, Subscript, Variable
from loopy import Reduction


@symbol
def type_localbasis_cache(element):
    return "LocalBasisCacheWithoutReferences<{}>".format(type_gfs(element))


@class_member("operator")
def define_localbasis_cache(element, name):
    include_file("dune/perftool/localbasiscache.hh", filetag="operatorfile")
    t = type_localbasis_cache(element)
    return "{} {};".format(t, name)


@symbol
def name_localbasis_cache(element):
    name = "cache_{}".format(FEM_name_mangling(element))
    define_localbasis_cache(element, name)
    return name


@preamble
def define_lfs_bound(lfs, bound):
    return 'auto {} = {}.size();'.format(bound, lfs)


@symbol
def name_lfs_bound(lfs):
    # LFS might either be given by an UFL element or by a string describig its name
    from ufl import FiniteElementBase
    if isinstance(lfs, FiniteElementBase):
        return name_lfs_bound(name_lfs(lfs))

    bound = '{}_size'.format(lfs)
    define_lfs_bound(lfs, bound)
    return bound


@preamble
def using_indices():
    return "using namespace Dune::Indices;"


@preamble
def define_lfs(name, father, child):
    using_indices()
    return "auto {} = child({}, _{});".format(name, father, child)


def lfs_child(lfs, child):
    return "{}.child({})".format(lfs, child)


@generator_factory(cache_key_generator=lambda e, r, **kw: (e, r))
def name_leaf_lfs(leaf_element, restriction, val=None):
    """ This function just caches leaf lfs names based on the
    element. The resulting local function spaces are useful only
    for size information. OTOH, they are available with just the
    leaf element available (as seen in basis evaluation).
    """
    # This generator function should be prepoluted by the lfs tree traversal,
    # so val should always be None when we actually want the result
    assert val

    return val


@generator_factory(cache_key_generator=lambda e, r, c, **kw: (e, r, c))
def name_lfs(element, restriction, component, prefix=None):
    # Omitting the prefix is only valid upon a second call, which will
    # result in a cache hit.
    assert prefix

    def _name_lfs(prefix, component):
        name = prefix
        if len(component) > 0:
            name = name + '_' + '_'.join(str(i) for i in component)
        return name

    name = _name_lfs(prefix, component)
    if len(component) > 0:
        father = _name_lfs(prefix, component[:-1])
        # If this localfunction space is the child of another one, trigger
        # the extraction preamble. Necessary before going into recursion
        # for having the correct (top-down) order of preambles
        define_lfs(name, father, component[-1])

    # Recurse into the given element to define all other local function spaces!
    from ufl import MixedElement
    from ufl.functionview import select_subelement
    from ufl.classes import FixedIndex
    subel = select_subelement(element, component)
    if isinstance(subel, MixedElement):
        for i in range(subel.num_sub_elements()):
            name_lfs(element, restriction, component + (FixedIndex(i),), prefix=prefix)

    # Cache the name for the subelement
    name_leaf_lfs(subel, restriction, val=name)

    # Now return the prefix!
    return name


@generator_factory(cache_key_generator=lambda e, **kw: e)
def type_gfs(element, basetype=None, index_stack=None):
    # Omitting basetype and index_stack is only valid upon a second call,
    # which will result in a cache hit.
    assert basetype
    assert index_stack is not None

    # Additionally, element is expected to be a ufl finite element
    from ufl import FiniteElementBase
    assert isinstance(element, FiniteElementBase)

    # Recurse into the given element to define all other local function spaces!
    from ufl import MixedElement
    from ufl.classes import FixedIndex
    if isinstance(element, MixedElement):
        for i, subelem in enumerate(element.sub_elements()):
            type_gfs(subelem, basetype=basetype, index_stack=index_stack + (FixedIndex(i),))

    if len(index_stack) == 0:
        return basetype
    else:
        include_file("dune/typetree/childextraction.hh", filetag="operatorfile")
        return 'Dune::TypeTree::Child<{},{}>'.format(basetype, ','.join(str(i) for i in index_stack))


def traverse_lfs_tree(arg):
    from dune.perftool.ufl.modified_terminals import ModifiedArgumentDescriptor
    assert isinstance(arg, ModifiedArgumentDescriptor)

    # First we need to determine the basename as given in the signature of
    # this kernel method!
    from dune.perftool.pdelab.argument import name_argumentspace
    lfs_basename = name_argumentspace(arg)
    from dune.perftool.pdelab.localoperator import lop_template_gfs
    gfs_basename = lop_template_gfs(arg)[1]

    # Now start recursively extracting local function spaces and fill the cache with
    # all those values. That way we can later get a correct local function space with
    # just the ufl finite element.
    from ufl.classes import MultiIndex
    name_lfs(arg.argexpr.ufl_element(), arg.restriction, MultiIndex(()), prefix=lfs_basename)
    type_gfs(arg.argexpr.ufl_element(), basetype=gfs_basename, index_stack=())


@generator_factory(item_tags=("iname",), cache_key_generator=lambda e, r, c: (e, c))
def _lfs_iname(element, restriction, context):
    lfs = name_leaf_lfs(element, restriction)
    bound = name_lfs_bound(lfs)

    name = "{}_{}_index".format(lfs, context)
    domain(name, bound)

    return name


def lfs_iname(element, restriction, count=None, context=''):
    """ Get the iname to iterate over the local function space given by element

    Arguments:
    ----------
    element: ufl.FiniteElementBase
        The finite element this local function space belongs to
    argcount: int
        Use to realize double nesting in case of jacobians
    context: str
        Some generation methods will require you to duplicate an iname for
        a given purpose, see the 'Loops and dependencies' of the loopy docs:
        https://documen.tician.de/loopy/tutorial.html#loops-and-dependencies
    """
    assert not ((context == '') and (count is None))
    if count is not None:
        if context != '':
            context = "{}_{}".format(count, context)
        else:
            context = str(count)
    return _lfs_iname(element, restriction, context)


@preamble
def declare_cache_temporary(element, restriction, name, which):
    t_cache = type_localbasis_cache(element)
    lfs = name_leaf_lfs(element, restriction)
    return "typename {}::{}ReturnType {};".format(t_cache,
                                                  which,
                                                  name,
                                                  )


@cached
def evaluate_basis(leaf_element, name, restriction):
    lfs = name_leaf_lfs(leaf_element, restriction)
    temporary_variable(name, shape=(name_lfs_bound(lfs),), decl_method=None)
    declare_cache_temporary(leaf_element, restriction, name, which='Function')
    cache = name_localbasis_cache(leaf_element)
    qp = name_quadrature_position_in_cell(restriction)
    instruction(inames=(quadrature_iname(),
                        ),
                code='{} = {}.evaluateFunction({}, {}.finiteElement().localBasis());'.format(name,
                                                                                             cache,
                                                                                             qp,
                                                                                             lfs,
                                                                                             ),
                assignees=frozenset({name}),
                read_variables=frozenset({qp}),
                )


@symbol
def name_basis(leaf_element, restriction):
    assert leaf_element.num_sub_elements() == 0
    # TODO name mangling!
    name = "phi_{}".format(FEM_name_mangling(leaf_element))
    name = restricted_name(name, restriction)
    evaluate_basis(leaf_element, name, restriction)
    return name


@cached
def evaluate_reference_gradient(leaf_element, name, restriction):
    lfs = name_leaf_lfs(leaf_element, restriction)
    temporary_variable(name, shape=(name_lfs_bound(lfs), name_dimension()), decl_method=None)
    declare_cache_temporary(leaf_element, restriction, name, which='Jacobian')
    cache = name_localbasis_cache(leaf_element)
    qp = name_quadrature_position_in_cell(restriction)
    instruction(inames=(quadrature_iname(),
                        ),
                code='{} = {}.evaluateJacobian({}, {}.finiteElement().localBasis());'.format(name,
                                                                                             cache,
                                                                                             qp,
                                                                                             lfs,
                                                                                             ),
                assignees=frozenset({name}),
                read_variables=frozenset({qp}),
                )


@symbol
def name_reference_gradient(leaf_element, restriction):
    assert leaf_element.num_sub_elements() == 0
    name = "js_{}".format(FEM_name_mangling(leaf_element))
    name = restricted_name(name, restriction)
    evaluate_reference_gradient(leaf_element, name, restriction)
    return name


@cached
def evaluate_basis_gradient(leaf_element, name, restriction):
    lfs = name_leaf_lfs(leaf_element, restriction)
    temporary_variable(name, shape=(name_lfs_bound(lfs), name_dimension()), shape_impl=('vec', 'fv'))
    jac = name_jacobian_inverse_transposed(restriction)
    index = lfs_iname(leaf_element, restriction, context='transformgrads')
    reference_gradients = name_reference_gradient(leaf_element, restriction)
    instruction(inames=(index,
                        quadrature_iname(),
                        ),
                code='{}.mv({}[{}][0], {}[{}]);'.format(jac,
                                                        reference_gradients,
                                                        index,
                                                        name,
                                                        index,
                                                        ),
                assignees=name,
                read_variables=frozenset({reference_gradients}),
                )


@symbol
def name_basis_gradient(leaf_element, restriction):
    assert leaf_element.num_sub_elements() == 0
    name = "gradphi_{}".format(FEM_name_mangling(leaf_element))
    name = restricted_name(name, restriction)
    evaluate_basis_gradient(leaf_element, name, restriction)
    return name


@cached
def evaluate_trialfunction(element, name, restriction, component):
    from ufl.functionview import select_subelement
    sub_element = select_subelement(element, component)

    # Determine the rank of the trialfunction tensor
    rank = len(sub_element.value_shape())
    assert rank in (0, 1)

    shape = (name_dimension(),) * rank
    shape_impl = ('fv', ) * rank
    idims = tuple(dimension_iname(count=i) for i in range(rank))
    leaf_element = sub_element
    from ufl import VectorElement
    if isinstance(sub_element, VectorElement):
        leaf_element = sub_element.sub_elements()[0]

    temporary_variable(name, shape=shape, shape_impl=shape_impl)
    lfs = name_lfs(element, restriction, component)
    index = lfs_iname(leaf_element, restriction, context='trial')
    basis = name_basis(leaf_element, restriction)
    assignee = Variable(name)

    if isinstance(sub_element, VectorElement):
        lfs = lfs_child(lfs, idims[0])
        assignee = Subscript(assignee, Variable(idims[0]))

    from dune.perftool.pdelab.argument import pymbolic_coefficient
    coeff = pymbolic_coefficient(lfs, index, restriction)
    reduction_expr = Product((coeff, Subscript(Variable(basis), Variable(index))))
    instruction(expression=Reduction("sum", index, reduction_expr, allow_simultaneous=True),
                assignee=assignee,
                forced_iname_deps=frozenset({quadrature_iname()}).union(frozenset(idims)),
                forced_iname_deps_is_final=True,
                )


@cached
def evaluate_trialfunction_gradient(element, name, restriction, component):
    # First we determine the rank of the tensor we are talking about
    from ufl.functionview import select_subelement
    sub_element = select_subelement(element, component)
    rank = len(sub_element.value_shape()) + 1
    assert rank in (1, 2)

    # We do then set some variables accordingly
    shape = (name_dimension(),) * rank
    if rank == 1:
        shape_impl = ('fv',)
    else:
        shape_impl = ('fm',)
    idims = tuple(dimension_iname(count=i) for i in range(rank))
    leaf_element = sub_element
    from ufl import VectorElement
    if isinstance(sub_element, VectorElement):
        leaf_element = sub_element.sub_elements()[0]

    # and proceed to call the necessary generator functions
    temporary_variable(name, shape=shape, shape_impl=shape_impl)
    lfs = name_lfs(element, restriction, component)
    index = lfs_iname(leaf_element, restriction, context='trialgrad')
    basis = name_basis_gradient(leaf_element, restriction)

    if isinstance(sub_element, VectorElement):
        lfs = lfs_child(lfs, idims[0])

    from dune.perftool.pdelab.argument import pymbolic_coefficient
    coeff = pymbolic_coefficient(lfs, index, restriction)
    assignee = Subscript(Variable(name), tuple(Variable(i) for i in idims))

    reduction_expr = Product((coeff, Subscript(Variable(basis), (Variable(index), Variable(idims[-1])))))
    instruction(expression=Reduction("sum", index, reduction_expr, allow_simultaneous=True),
                assignee=assignee,
                forced_iname_deps=frozenset({quadrature_iname()}).union(frozenset(idims)),
                forced_iname_deps_is_final=True,
                )
